/*!
  \file   Example2.C
  \brief  Polynome de Chaos - Cubature - ANOVA
*/
#include <iostream>

// Les classes Nisp
#include "nisp_pc.h"

using namespace std;

// Modele mathematique : y(0) = x1 * x2 et y(1) = 2 * x1 * x2
void modele(double x[], double y[]) {
  y[0] =      x[0] * x[1];
  y[1] = 2. * x[0] * x[1];
}

// Edition de l'analyse
void EditionAnalyse(PolynomialChaos * pc) {
  int nx = pc->GetDimensionInput();
  int ny = pc->GetDimensionOutput();

  for(int j=1;j<=ny;j++) {
    cout << "Variable : " << j << endl;
    cout << "=============" << endl;
    cout << "Moyenne  : " << pc->GetMean(j)              << endl;
    cout << "Variance : " << pc->GetVariance(j)          << endl;
    cout << endl;

    cout << "Indice sensibilite du premier ordre"<<endl;
    for(int i=1;i<=nx;i++) {
      cout << "  Variable ["<<i<<"]=" << pc->GetIndiceFirstOrder(i) << endl;
    }
    cout << endl;

    cout << "Indice sensibilite totale"<<endl;
    for(int i=1;i<=nx;i++) {
      cout << "  Variable ["<<i<<"]=" << pc->GetIndiceTotalOrder(i) << endl;
    }
    cout << endl;

    // Indice de sensibilite d'un groupe de variables
    cout << "Fraction de la variance d'un groupe de variables"<<endl;
    pc->SetGroupEmpty();
    pc->SetGroupAddVar(1);
    cout << "  Groupe [1]=" << pc->GetGroupIndice(j) << endl;
    pc->SetGroupEmpty();
    pc->SetGroupAddVar(2);
    cout << "  Groupe [2]=" << pc->GetGroupIndice(j) << endl;
    pc->SetGroupEmpty();
    pc->SetGroupAddVar(1);
    pc->SetGroupAddVar(2);
    cout << "  Groupe [1,2]=" << pc->GetGroupIndice(j) << endl;
    cout << endl;

    cout << "Fraction de la variance d'un groupe de variables en interaction"<<endl;
    pc->SetGroupEmpty();
    pc->SetGroupAddVar(1);
    cout << "  Groupe [1]=" << pc->GetGroupIndiceInteraction(j) << endl;
    pc->SetGroupEmpty();
    pc->SetGroupAddVar(2);
    cout << "  Groupe [2]=" << pc->GetGroupIndiceInteraction(j) << endl;
    pc->SetGroupEmpty();
    pc->SetGroupAddVar(1);
    pc->SetGroupAddVar(2);
    cout << "  Groupe [1,2]=" << pc->GetGroupIndiceInteraction(j) << endl;
    cout << endl;

    // Pour afficher toute la decomposition fonctionnelle de la variance
    cout << "ANOVA fonctionnelle"<< endl;
    pc->GetAnova(j);
    cout << "ANOVA fonctionnelle ordonnee"  << endl;
    pc->GetAnovaOrdered(j);
    cout << "ANOVA fonctionnelle ordonnee et seuillee a 95%"  << endl;
    pc->GetAnovaOrdered((double) 0.95,j);
    cout << endl;
  }
}

/*!
  \fn void Example2(int degre)
  \brief Polynome de Chaos - Quadrature - Indices de Sobol - ANOVA
  \param degre : degre maximal du polynome
*/
void Example2(int degre) {

  cout.setf(ios::scientific, ios::floatfield);
  cout.precision(8);

  // Definition des variables physiques incertaines
  SetRandomVariable * gu = new SetRandomVariable();             // groupe de variables incertaines
  gu->AddRandomVariable(new RandomVariable("Normale",1.,1.));   // moyenne 1., ecart-type 1
  gu->AddRandomVariable(new RandomVariable("Normale",1.,4.));   // moyenne 2., ecrat-type 4

  // Pour editer des informations sur la sortie standard
  cout << "Informations sur le groupe de parametres incertains"<<endl;
  gu->GetLog();

  // Definition du groupe de variables stochastiques : uniformes [0,1]
  int nx = gu->GetDimension();  // nx=2
  SetRandomVariable * gx = new SetRandomVariable(nx);

  // Pour editer des informations sur la sortie standard
  cout << "Informations sur le groupe des variables stochastiques"<<endl;
  gx->GetLog();

  // Calcul du plan d'experiences numeriques par Cubature
  gx->BuildSample("Petras",degre);
  gu->BuildSample(gx);

  // Sauvegarde des plans d'experiences
  gx->Save((char *) "PlanStochastique.dat");
  gu->Save((char *) "PlanExperiences.dat");

  // Polynome de chaos
  int ny=2; // polynomes a 2 sorties
  PolynomialChaos * pc = new PolynomialChaos(gx,ny); // legendre x legendre
  //int ny = pc->GetDimensionOutput(); // ny=1 (valeur par defaut lors de la construction dun PC)

  // Pour editer les informations
  cout << "Informations sur le polynome"<<endl;
  pc->GetLog();

  // Realisation du plan d'experiences numeriques
  int np=gu->GetSize();  // taille du plan d'experiences
  pc->SetSizeTarget(np); // allocation memoire 
  double input[nx], output[ny];
  for(int k=1;k<=np;k++) {
    for(int i=0;i<nx;i++) input[i]=gu->GetSample(k,i+1);  // on recupere les entrees
    modele(input,output);                               // appel au modele numerique
    for(int j=0;j<ny;j++) pc->SetTarget(k,j+1,output[j]); // on transmet au PC les sorties calculees
  }

  // Calcul du chaos polynomial
  pc->SetDegree(degre); // on pourrait specifier un degre inferieur
  pc->ComputeChaosExpansion(gx,"Integration");

  // Sauvegarde du polynome
  pc->Save((char *) "NispExample2.dat");

  // Generation du code source
  pc->GenerateCode((char *) "NispExample2.C", (char *) "NispExample2");

  // Edition de l'analyse 
  EditionAnalyse(pc);
}

int main(int argc, char *argv[]) {
  if(argc != 2) {
    cout << "Erreur !! la commande est de la forme << nisp degre >>" << endl;
    cout << "degre : degre du polynome" << endl;
  }
  else {
    int degre = atoi(argv[1]);
    Example2(degre);
  }
}
