/*!
  \file   Ishigami.C
  \brief  Polynomes de Chaos - Analyse de Sensibilite
*/
#include <iostream>
#include <cmath>     // pour disposer de la macro M_PI = 3.14159...
#include "nisp_pc.h"   // les classes Nisp
#include "nisp_fun.h" // les utilitaires de Nisp

using namespace std;

// Fonction de Ishigami
// Modele y1(x1,x2,x3) = sin(x1) + a (sin(x2))**2 + b (x3)**4 sin(x1)
// Modele y2(x1,x2,x3) = 2 * [sin(x1) + a (sin(x2))**2 + b (x3)**4 sin(x1)]
void Ishigami(double x[], double y[]) {
  double u1,u2,u3,s1,s2,s3,a,b;
  a=7.; b=0.1;
  u1=x[0];    u2=x[1];    u3=x[2];
  s1=sin(u1); s2=sin(u2); s3=sin(u3);
  y[0] = s1 + a *s2*s2 + b * u3*u3*u3*u3 * s1;
  y[1]=2.*y[0];
}


/*!
  \fn void CaseIshigami(char * methode, int degre)
  \brief Polynomes de Chaos par quadrature - cubature
  \param methode : methode de quadrature ("Quadrature") ou de cubature ("Petras")
  \param degre   : degre du polynome
*/
void CaseIshigami(char *methode, int degre) {
  SetRandomVariable * gu = new SetRandomVariable();
  gu->AddRandomVariable(new RandomVariable("Uniforme",-M_PI,+M_PI));
  gu->AddRandomVariable(new RandomVariable("Uniforme",-M_PI,+M_PI));
  gu->AddRandomVariable(new RandomVariable("Uniforme",-M_PI,+M_PI));

  gu->GetLog();

  int nx=gu->GetDimension(); // nx=3
  SetRandomVariable * gx = new SetRandomVariable(nx);
  gx->BuildSample(methode,degre);

  // Nombre d'evaluations numeriques de la fonction Ishigami
  cout << "Nombre de simulations = " << gx->GetSize() << endl;

  // Plan d'experiences
  gu->BuildSample(gx);

  // Sauvegarde des plans d'experiences
  gx->Save((char *) "PlanStochastique.dat");
  gu->Save((char *) "PlanExperiences.dat");

  int ny=2; // nombre de sorties
  PolynomialChaos * pc = new PolynomialChaos(gx,ny);

  // Realisation du plan d'experiences
  int np=gu->GetSize();
  pc->SetSizeTarget(np);
  int ne = pc->GetDimensionInput();
  int ns = pc->GetDimensionOutput();
  double input[ne], output[ns]; // variables temporaires
  for(int k=1;k<=np;k++) {
    for(int i=1;i<=ne;i++) input[i-1]=gu->GetSample(k,i);  // on recupere les entrees
    Ishigami(input,output);                                // on realise l'experience numerique
    for(int j=1;j<=ns;j++) pc->SetTarget(k,j,output[j-1]); // on transmet les sorties calculees au PC
  }

  // Calcul du PC
  pc->SetDegree(degre);
  pc->ComputeChaosExpansion(gx,"Integration");

  // Sauvegarde du polynome
  pc->Save((char *) "NispIshigami.dat");
  // Generation du code source
  pc->GenerateCode((char *) "NispIshigami.C", (char *) "NispIshigami");

  // Moyenne et variance
  cout << "Moyenne"<<endl;
  for(int j=1;j<=ns;j++) cout << "  Moyenne["  << j << "] =" << pc->GetMean(j)    << endl;
  cout << "Variance"<<endl;
  for(int j=1;j<=ns;j++) cout << "  Variance[" << j << "]=" << pc->GetVariance(j) << endl;

  // Matrices des covariances et correlations
  cout << "MATRICE Covariances" << endl;
  for(int i=1;i<=ns;i++) {
    for(int j=1;j<=ns;j++) cout << pc->GetCovariance(i,j) << " ";
    cout << endl;
  }
  cout << "MATRICE Correlations" << endl;
  for(int i=1;i<=ns;i++) {
    for(int j=1;j<=ns;j++) cout << pc->GetCorrelation(i,j) << " ";
    cout << endl;
  }
  // Indices de sensibilite
  for(int j=1;j<=ns;j++) {
    cout << "Indice de sensibilite - Output[" << j << "]"<<endl;
    cout << "  Premier ordre ***"<< endl;
    for(int i=1;i<=ne;i++) cout << "  Variable[" << i << "]= " << pc->GetIndiceFirstOrder(i,j) << endl;
    cout << "  Totale **********"<< endl;
    for(int i=1;i<=ne;i++) cout << "  Variable[" << i << "]= " << pc->GetIndiceTotalOrder(i,j) << endl;
  }

  // Pour afficher l'ANOVA fonctionnelle pour les 2 variables de sortie
  for(int j=1;j<=ns;j++) {
    cout << "Output " << j << " ****"<<endl;
    cout << "ANOVA fonctionnelle"<< endl;
    pc->GetAnova(j);
    cout << "ANOVA fonctionnelle ordonnee"  << endl;
    pc->GetAnovaOrdered(j);
    cout << "ANOVA fonctionnelle ordonnee et seuillee a 95%"  << endl;
    pc->GetAnovaOrdered((double) 0.95,j);
  }

  // Indices de sensibilite d'un groupe de varibales
  cout << "GROUPE de VARIABLES" << endl;
  for(int j=1;j<=ns;j++) {
    pc->SetGroupEmpty();
    pc->SetGroupAddVar(1);
    pc->SetGroupAddVar(3);
    cout << "Output " << j << " ****"<<endl;
    cout << "   Indice du groupe (1,3) " << pc->GetGroupIndice(j) << endl;
    cout << "   Indice du groupe (1,3) Interaction " << pc->GetGroupIndiceInteraction(j) << endl;
  }

  // Quantiles a 95%
  cout << "QUANTILE" << endl;
  InitSeed(1);
  pc->BuildSample("MonteCarlo",10000);
  cout << pc->GetQuantile((double) 0.95) << endl;
  cout << pc->GetInvQuantile(pc->GetQuantile((double) 0.95)) << endl;
}


void ValeursExactes() {
// Valeurs exactes
  double moy;	// 3.5
  double var;	// 13.8445
  double s1;	// 0.3139	indices[1]
  double s2;	// 0.4424	indices[2]
  double s12;	// 0. 		indices[3]
  double s3;	// 0.		indices[4]
  double s13;	// 0.2437	indices[5]
  double s23;	// 0.		indices[6]
  double s123;	// 0.		indices[7]

  // Calcul des valeurs exactes
  double a,b,d1,d2,d13;
  a=7.; b=0.1;
  d1 =(b*pow(M_PI,4.))/5 + b*b*pow(M_PI,8.)/50. + 0.5;
  d2 = a*a/8;
  d13=pow(M_PI,8.)*b*b*(1./18-1/50.);
  moy=a/2.;
  var=d1+d2+d13;
  s1 = d1/var; s2 = d2/var; s13=d13/var;
  s3=s12=s23=s123=0.;
}

int main(int argc, char *argv[]) {
  if(argc != 3) {
    cout << "Commande est de la forme << nisp methode degre >>" << endl;
    cout << "1 - methode : methode de quadrature(Quadrature)/cubature(Petras)" << endl;
    cout << "2 - degre   : degre du polynome" << endl;
  }
  else {
    int degre  = atoi(argv[2]);
    CaseIshigami(argv[1],degre);
  }
}
