/*!
  \file   RegressionIshigami.C
  \brief  Polynomes de Chaos - Regression - Analyse de Sensibilite
*/
#include <iostream>
#include <cmath>     // pour disposer de la macro M_PI = 3.14159...
#include "nisp_pc.h" // les classes Nisp

using namespace std;

// Fonction de Ishigami
// Modele y1(x1,x2,x3) = sin(x1) + a (sin(x2))**2 + b (x3)**4 sin(x1)
// Modele y2(x1,x2,x3) = 2 * [sin(x1) + a (sin(x2))**2 + b (x3)**4 sin(x1)]
void Ishigami(double x[], double y[]) {
  double u1,u2,u3,s1,s2,s3,a,b;
  a=7.; b=0.1;
  u1=x[0];    u2=x[1];    u3=x[2];
  s1=sin(u1); s2=sin(u2); s3=sin(u3);
  y[0] = s1 + a *s2*s2 + b * u3*u3*u3*u3 * s1;
  y[1]=2.*y[0];
}


/*!
  \fn void RegressionIshigami(char * methode, int np)
  \brief Polynomes de Chaos par quadrature - cubature
  \param methode : methode d'echantillonnage (MonteCarlo, Lhs, QmcSobol)
  \param np      : nombre de simulations
*/
void RegressionIshigami(char *methode, int np) {

  SetRandomVariable * gu = new SetRandomVariable();
  gu->AddRandomVariable(new RandomVariable("Uniforme",-M_PI,+M_PI));
  gu->AddRandomVariable(new RandomVariable("Uniforme",-M_PI,+M_PI));
  gu->AddRandomVariable(new RandomVariable("Uniforme",-M_PI,+M_PI));

  int nx=gu->GetDimension();
  SetRandomVariable * gx = new SetRandomVariable(nx);
  gx->BuildSample(methode,np);

  // Plan d'experiences
  gu->BuildSample(gx);

  // Sauvegarde des plans d'experiences
  gx->Save((char *) "PlanStochastique.dat");
  gu->Save((char *) "PlanExperiences.dat");

  int ny=2; // nombre de sorties
  PolynomialChaos * pc = new PolynomialChaos(gx,ny);

  // Realisation du plan d'experiences
  pc->SetSizeTarget(np);
  double input[nx], output[ny]; // variables temporaires
  for(int k=0;k<np;k++) {
    for(int i=0;i<nx;i++) input[i]=gu->GetSample(k+1,i+1);  // on recupere les entrees
    Ishigami(input,output);                                 // on realise l'experience numerique
    for(int j=0;j<ny;j++) pc->SetTarget(k+1,j+1,output[j]); // on transmet les sorties calculees au PC
  }

  // Calcul du PC
  pc->SetDegree(6);
  pc->ComputeChaosExpansion(gx,"Regression");

  // Moyenne et variance
  cout << "Moyenne"<<endl;
  for(int j=0;j<ny;j++) cout << "  Moyenne["<<j+1<<"] =" << pc->GetMean(j+1) << endl;
  cout << "Variance"<<endl;
  for(int j=0;j<ny;j++) cout << "  Variance["<<j+1<<"]=" << pc->GetVariance(j+1) << endl;

  // Indices de sensibilite
  for(int j=0;j<ny;j++) {
    cout << "Indice de sensibilite - Output["<<j+1<<"]"<<endl;
    cout << "  Premier ordre ***"<<endl;
    for(int i=0;i<nx;i++) cout << "  Variable["<<i+1<< "]= " << pc->GetIndiceFirstOrder(i+1,j+1) << endl;
    cout << "  Totale **********"<<endl;
    for(int i=0;i<nx;i++) cout << "  Variable["<<i+1<< "]= " << pc->GetIndiceTotalOrder(i+1,j+1) << endl;
  }

  // Pour afficher l'ANOVA fonctionnelle pour les 2 variables de sortie
  for(int j=0;j<ny;j++) {
    cout << "Output "<<j+1<<"****"<<endl;
    cout << "ANOVA fonctionnelle"<< endl;
    pc->GetAnova(j+1);
    cout << "ANOVA fonctionnelle ordonnee"  << endl;
    pc->GetAnovaOrdered(j+1);
    cout << "ANOVA fonctionnelle ordonnee et seuillee a 95%"  << endl;
    pc->GetAnovaOrdered((double) 0.95,j+1);
  }
}

int main(int argc, char *argv[]) {
  if(argc != 3) {
    cout << "Commande est de la forme << nisp methode np >>" << endl;
    cout << "1 - methode : echantillonnage (MonteCarlo, Lhs, QmcSobol)" << endl;
    cout << "2 - np   : nombre de simulations" << endl;
  }
  else {
    int np  = atoi(argv[2]);
    RegressionIshigami(argv[1],np);
  }
}
