// 2006-2009 - CEA - Jean-Marc Martinez
// 2009 - INRIA - Michael Baudin
#include <iostream>
#include <sstream>
#include <fstream>
#include <vector>

#define _USE_MATH_DEFINES
#include "math.h"

#include "nisp_pc.h"

using namespace std;

#include "nisp_util.h"
#include "nisp_math.h"
#include "nisp_smolyak.h"
#include "nisp_msg.h"
#include "nisp_qua.h"

// Methode du trapeze (bords retires n = 2 ** (l-1) + 1)
void TrapezoidalRule(double x[], double w[], int n) {
  int i;
  double delta = 1./ (n + 1.);
  for(i=0;i<n;i++) {
	  x[i]= (i+1.)*delta; w[i]=1./n;
  }
}

// Methode de Fejer
void FejerRule(double x[], double w[], int n) {
  int i,j;
  double theta;
  for(i=0;i<n;i++) {
    theta=(i+1.)*M_PI/(n+1.);
    x[i]=cos(theta);
	for(w[i]=0,j=0;j<(n+1)/2;j++) {
		w[i]+=sin((2.*j+1.)*theta)/(2.*j+1.);
	}
    w[i]=w[i]*sin(theta)*4./(n+1.);
  }
  for(i=0;i<n;i++) {
	  x[i]=(x[i]+1.)/2.; 
	  w[i]=w[i]/2.;
  } // entre [0,1]
}

// Methode de Clenshaw Curtis
void ClenshawCurtisRule(double x[], double w[], int n) {
  int i,j;
  double theta;
  if(n==1) {x[0]=0.; w[0]=2.;}
  else { // ici n est > 2
	  for(i=0;i<n;i++) {
		  x[i]=-cos(M_PI * i / (n-1));
	  }
    w[0]=w[n-1]=1./(n*(n-2.));
    for(i=1;i<n-1;i++) {
      theta=M_PI * i / (n-1);
	  for(w[i]=0.,j=1;j<=(n-1)/2-1;j++) {
		  w[i]+=2./(1-4*j*j) * cos(2.*theta*j);
	  }
      w[i]+=1./(1-4*j*j) * cos(theta*(n-1));
      w[i]=(2./(n-1))*(1.+w[i]);
    }
  }
  for(i=0;i<n;i++) {
	  x[i]=(x[i]+1.)/2.; 
	  w[i]=w[i]/2.;
  } // entre [0,1]
}

class FormuleQuadrature {
public:
  string type;
  int niveau;
  int *np;    // np[0 ... niveau-1]
  double **x; // x[i=0 ... niveau-1][0,...,np[i]]
  double **w; // w[i=0 ... niveau-1][0,...,np[i]]
public:
  FormuleQuadrature(string type_va, string type_quadrature, int level);
  ~FormuleQuadrature();
};

FormuleQuadrature::FormuleQuadrature(string type_va, string type_quadrature, int level) {
  int i,j;
  type=type_quadrature;
  niveau=level;
  np=ivector(level);
  if(type_quadrature=="SmolyakTrapeze") {   // np = 1 3 7 15 ... n = 2**level - 1
    np[0]=1;
	for(i=1;i<level;i++) {
		np[i]=2*np[i-1]+1;
	}
  }
  else if(type_quadrature=="SmolyakFejer") { // np = 1 3 7 15 ... n = 2**level - 1
    np[0]=1;
    // TODO : check these nested loops
    for(i=1;i<level;i++) {
		for(np[i]=1,j=0;j<=i;j++) {
			np[i]=np[i]*2;
		}
		np[i]=np[i]-1;
	}
  }
  else if(type_quadrature=="SmolyakClenshawCurtis") { // np = 1 3 5 9 ... n = 2**(level-1) + 1 si level >=2
    np[0]=1;
    // TODO : check these nested loops
    for(i=1;i<level;i++) {
		for(np[i]=1,j=0;j<i;j++) {
			np[i]=np[i]*2;
		}
		np[i]=np[i]+1;
	}
  }
  else {
	  for(i=0;i<level;i++) {
		  np[i]=i+1; // Gauss
	  }
  }

  x = new double * [level];
  w = new double * [level];
  for(i=0;i<level;i++) {
	  x[i] = new double[np[i]];
  }
  for(i=0;i<level;i++) {
	  w[i] = new double[np[i]];
  }
  
  if(type_quadrature=="SmolyakTrapeze")    {
	  for(i=0;i<level;i++) {
		  TrapezoidalRule(x[i],w[i],np[i]);
	  }
  }
  else if(type_quadrature=="SmolyakFejer") {
	  for(i=0;i<level;i++) {
		  FejerRule(x[i],w[i],np[i]);
	  }
  }
  else if(type_quadrature=="SmolyakClenshawCurtis") {
	  for(i=0;i<level;i++) {
		  ClenshawCurtisRule(x[i],w[i],np[i]);
	  }
  }
  else {
	  for(i=0;i<level;i++) {
		  Quadrature0(x[i],w[i],np[i],type_va); // Gauss
	  }
  }
}

FormuleQuadrature::~FormuleQuadrature() {
  for(int i=0;i<niveau;i++) {
	  delete[] x[i]; 
	  delete[] w[i];
  }
  free_ivector(np);
}

// incrementation du multi indice (k(0),k(1),...,k(nx-1))
// min(i)<= k(i) <= max(i)
// Remarque : avant l'appel l'initialisation doit avoir ete effectuee
// Valeur initiale du multi indice (min(0),min(1),...,min(nx-1))
// retour 1 si l'incrementation a ete realisee 
// retour 0 si l'incrementation n'a pas ete realisee
// car le maximun a ete atteint (max(0),max(1),...,max(nx-1)
int incrementation(int k[], int min[], int max[], int nx) {
  for(int i=0;i<nx;i++) {
    if(k[i]<max[i]) {
      k[i]+=1;
	  for(int j=0;j<i;j++) {
		  k[j]=min[j];
	  }
      return(1);
    }
  }
  return(0);
}

// Norme L1 du multi indice
int norme_L1(int k[], int nx) {
  int i,s=0;
  for(i=0;i<nx;i++) s+=k[i];
  return(s);
}

//Norme Linfini du multi indice
int norme_Linf(int k[], int nx) {
  int i,s;
  for(s=k[0],i=1;i<nx;i++) {
	  if(k[i]>s) {
		  s=k[i];
	  }
  }
  return(s);
}

// Filtre : norme min <= L1 <= max
int Filtre_norme_L1(int k[], int nx, int vmin, int vmax) {
  int norme=norme_L1(k,nx);
  if((norme>=vmin) && (norme<=vmax)) {
	  return(1);
  }
  else {
	  return(0);
  }
}

int Cnp(int n, int p) {
	if(n<p) {
		cout << " n < p !!" << endl;
	}
	if(p<0) {
		cout << " p < 0 !!" << endl;
	}
  int i,fn,fd;
  // factoriel n sur factoriel p
  for(fn=1,i=p+1;i<=n;i++) {
	  fn=fn*i;
  }
  // factoriel n-p
  for(fd=1,i=1;i<=(n-p);i++) {
	  fd=fd*i;
  }
  return(fn/fd);
}

// Constition de l'ensemble des multi indices
// A l'etat initial *indices est un vecteur nul
// A l'etat final *indices contient tous les multi indices k
// qui verifient la contrainte vmin <= norme_L1(k) <= vmax
void Build_MultiIndex_Filtre_norme_L1(vector<int *> * indices, int * min, int * max, int nx, int vmin, int vmax) {
  int * k = ivector(nx);
  int * knew;
  int i;

  for(i=0;i<nx;i++) k[i]=min[i];
  if(Filtre_norme_L1(k,nx,vmin,vmax)) {
    knew = ivector(nx);
	for(i=0;i<nx;i++) {
		knew[i]=k[i];
	}
    indices->push_back(knew);
  }
  while (incrementation(k,min,max,nx)) {
    if(Filtre_norme_L1(k,nx,vmin,vmax)) {
      knew = ivector(nx);
	  for(i=0;i<nx;i++) {
		  knew[i]=k[i];
	  }
      indices->push_back(knew);
    }
  }
  free_ivector(k);
}

void Compute_Coef_Cnp_Signe(int *coef_Cnp, int *signe, vector<int *> * indices, int nx, int l) {
  vector<int *>::iterator it;
  int norme_L1(int *k, int nx);
  int *k, norme, j=0;
  for(it=indices->begin();it!=indices->end();++it) {
    k=*it;
    norme = norme_L1(k,nx);
    coef_Cnp[j]= Cnp(nx-1,norme-l);
	if((l+nx-norme-1)%2) {
		signe[j]=-1;
	}
	else {
		signe[j]=1;
	}
    j++;
  }
}

// Formule est exacte au degre 2*l-1
// On peut donc construire des polynomes de degre l-1
// Remarque : on peut integrer des polynomes de dege > 2*l-1 en faisant
// une legere erreur.
// Par exemple n=9 l=3, l'integrale de x**8 dans [0,1] = 0.11111 et on
// trouve 0.108458 et donc une erreur de l'ordre de 3%
//
//
// type = {"SmolyakTrapeze", "SmolyakGauss", "SmolyakFejer"}
// Pour Fejer : le niveau est degre + 1
//            : si on veut calculer un PC de degre 5 => niveau 6
//            : REMARQUE : si la vraie fonction est un polynome de degre d+1
//                         le calcul des coefficients du PC de degre d est exacte
// Pour Gauss : idem 
// IMPORTANT : ces regles sont penalisantes lorsque la dimension
//             est petite. Par exemple en prenant nx=2 Fejer de niveau 7
//             permet de calculer un PC de degre 11 > 6 !! 
//
// Methodologie : pour nx < 3 utiliser la quadrature tensorisee
//
void smolyak(SetRandomVariable * gx, string type, int degre) {
  int nx=gx->nx;
  int l;
  if(type=="SmolyakGauss") {
	  l=degre+1; // comme pour Quadrature tensorisee
  }
  else {
	  l=degre + 1; // la meme regle pour Fejer (voir ci dessous)
  }
  // si nx==2 : si degre<=5  l=degre+1
  //               degre<=7  l=6
  //               degre<=11 l=7
  //               degre<=15 l=8
  //               degre<=23 l=9
  // si nx==3 : si degre<=5  l=degre+1
  //               degre<=7  l=7
  //               degre==8  l=8
  //               degre==9  l=8
  //               degre==10 l=9
  // si nx==4  on prend la regle l=degre+1

  if(nx<1) {
    ostringstream msg;
    msg<<"NISP - ERROR"<<endl;
    msg << "Nisp(smolyak) : stochastic dimension = " << nx << " < 1" << endl;
    nisp_error ( msg.str() );
    return;
  }
  if(degre<0) {
    ostringstream msg;
    msg<<"NISP - ERROR"<<endl;
    msg << "Nisp(smolyak) : degree  == " << degre << " < 0" << endl;
    nisp_error ( msg.str() );
    return;
  }
  // Formule de quadrature a 1 dimension
  //FormuleQuadrature * fq[nx];
  vector<FormuleQuadrature *> fq(nx);
  for(int i=0;i<nx;i++) {
		fq[i]=new FormuleQuadrature(gx->va[i]->type,type,l);
  }

  // Points et poids de la sparse grid
  // Ils seront ensuite recopies dans le champ (x et w) dans gx (SetRandomVariable)
  vector<double *> * point  = new vector<double *>();
  vector<double>   * weight = new vector<double>();

  // Creation des multi indices dont la norme L1
  // est comprise entre l et l+nx-1 (bornes incluses)
  vector<int *>   * indices = new vector<int *>();
  int * min = ivector(nx);
  int * max = ivector(nx);
  for(int i=0;i<nx;i++) {
	  min[i]=1;
  }
  for(int i=0;i<nx;i++) {
	  max[i]=l;
  }
  Build_MultiIndex_Filtre_norme_L1(indices,min,max,nx,l,l+nx-1);

  int * coef_Cnp = ivector(indices->size());
  int * signe    = ivector(indices->size());
  Compute_Coef_Cnp_Signe(coef_Cnp,signe,indices,nx,l);

  vector<int *>::iterator it;
  int *k,i,j=0;
  int * nmin = ivector(nx); // premier point : en general 1
  int * nmax = ivector(nx); // dernier point : en general egal au nombre maximal de points de
                            // la quadrature a une dimension
  int * kc   = ivector(nx);

  double * xc = dvector(nx);
  double * xcc;
  double poids;
  for(i=0;i<nx;i++) nmin[i]=1;
  for(it=indices->begin();it!=indices->end();++it) {
    k=*it; // c'est k qui determine la tensorisation
    // initialisation du multi indice courant
    for(i=0;i<nx;i++) 
	{
		kc[i]=nmin[i];
	}
    // Premier point ...
    //cout<<"Quadrature : ";for(i=0;i<nx;i++) cout<<k[i]<<" ";cout<<endl;
    //for(i=0;i<nx;i++) cout << x[k[i]-1][0] << " ";cout << endl;
    for(poids=coef_Cnp[j]*signe[j],i=0;i<nx;i++) 
	{
		poids=poids*fq[i]->w[k[i]-1][0];
	}
	for(i=0;i<nx;i++) {
		xc[i]=fq[i]->x[k[i]-1][0];
	}
    xcc = dvector(nx); for(i=0;i<nx;i++) 
		xcc[i]=xc[i]; point->push_back(xcc);
    weight->push_back(poids);

    // ... et points suivants eventuellement
    for(i=0;i<nx;i++) nmax[i]=fq[i]->np[k[i]-1];
    //for(i=0;i<nx;i++) cout << nmax[i] << " ";cout << endl;
    while(incrementation(kc,nmin,nmax,nx)) {
      //for(i=0;i<nx;i++) cout << x[k[i]-1][kc[i]-1] << " ";cout << endl;
      //for(i=0;i<nx;i++) cout << kc[i] << " ";cout << endl;
      for(poids=coef_Cnp[j]*signe[j],i=0;i<nx;i++) 
	  {
		  poids=poids*fq[i]->w[k[i]-1][kc[i]-1];
	  }
      for(i=0;i<nx;i++) 
	  {
		  xc[i]=fq[i]->x[k[i]-1][kc[i]-1];
	  }
      xcc = dvector(nx); 
	  for(i=0;i<nx;i++) 
	  {
		  xcc[i]=xc[i]; 
		  point->push_back(xcc);
	  }
      weight->push_back(poids);
    }
    j++; // l'indice pour coef_Cnp et signe
  }

  // On sauvegarde le plan stochastique dans gx (SetRandomVariable)
  int np=point->size();
  gx->typesampling=type;
  gx->degre_max=degre;
  gx->x=dmatrix(np+1,nx+1);
  gx->w=dvector(np+1);
  gx->np=np;
  int ki=0;
  double *xp,wp;
  vector<double *>::iterator ipoint;
  for(ipoint=point->begin();ipoint!=point->end();++ipoint) {
    xp=*ipoint;
    wp=(*weight)[ki];
    for(int i=0;i<nx;i++) 
	{
		gx->x[ki+1][i+1]=xp[i];
	}
    gx->w[ki+1]=wp;
    ki++;
  }

  // nettoyage
  free_dvector(xc);
  free_ivector(coef_Cnp);
  free_ivector(signe);
  free_ivector(min);
  free_ivector(max);
  for(it=indices->begin();it!=indices->end();++it) {
    k=*it;
    free_ivector(k);
  }
  
  // Algo pas du tout efficace !!!
  // Si formules emboitees on doit retirer les doublons
  if(type=="SmolyakGauss") return; // Formule non emboitees pour Gauss

  // Si doublon on recopie le poids dans celui qui lui correspond puis on le met a 0
  int k1, k2, flag;
  for(k1=1;k1<=np;k1++) {
	  if(gx->w[k1]) {
		  for(k2=k1+1;k2<=np;k2++) {
			  if(gx->w[k2]) {
				  flag=1;
				  //for(i=1;i<=nx;i++) if(gx->x[k1][i]!=gx->x[k2][i]) {flag=0;}
				  for(i=1;i<=nx;i++) 
				  {
					  if(fabs(gx->x[k1][i]-gx->x[k2][i])>1.e-6) {
						  flag=0;
						  i=nx+1;
					  }
				  }
				  if(flag) {
					  gx->w[k1]+=gx->w[k2];
					  gx->w[k2]=0.;
				  }
			  }
		  }
	  }
  }

  // Nombre de points differents (on ne compte pas ceux pour lesquels le poids = 0)
  int npr=0;
  for(k1=1;k1<=np;k1++) {
	  if(gx->w[k1]) {
		  npr++;
	  }
  }

  // On reconstitue l'echantillon
  double **xr = dmatrix(npr+1,nx+1);
  double *wr  = dvector(npr+1);
  int kt;
  for(kt=1,k1=1;k1<=np;k1++) {
    if(gx->w[k1]) {
		for(i=1;i<=nx;i++) {
			xr[kt][i]=gx->x[k1][i];
		}
      wr[kt]=gx->w[k1];
      kt++;
    }
  }
  free_dmatrix(gx->x,np+1);
  free_dvector(gx->w);
  gx->x=xr;
  gx->w=wr;
  gx->np=npr;
  for(int i=0;i<nx;i++) {
		delete fq[i];
  }
  delete point;
  delete weight;
  delete indices;
}

